@@ -3,8 +3,8 @@
 
 #define CELT_BUILD            1
 
-#define restrict 
 #define inline __inline
+#define getpid _getpid
 
 #define USE_ALLOCA            1
 
@@ -18,4 +18,6 @@
 #pragma warning(disable : 4996)/* This function or variable may be unsafe. Consider using fopen_s instead. To disable deprecation, use _CRT_SECURE_NO_WARNINGS. See online help for details. */
 #endif
 
+#include "version.h"
+
 #endif CONFIG_H
